@@ -39,6 +39,7 @@ future_dispatch_lock = asyncio.Lock()
 # mention_pattern = re.compile(r'<@.*?(\d+)>')
 
 FALLBACK_GENERAL = ''
+DEFAULT_NULL_GUILD_ID = 0
 
 
 class EventChain(object):
@@ -79,15 +80,14 @@ class Client(discord.Client):
     def __init__(self, control_interface, *args, intents=None, status=discord.Status.dnd, activity=discord.Game(name="Starting up, please wait..."), control_meta=None, **kwargs):
         if intents is None:
             intents = discord.Intents.default()
-        super().__init__(*args, intents=intents, status=status, activity=activity, **kwargs)
+        super().__init__(*args, intents=intents, status=status, activity=activity, shard_count=control_meta.get('shard_count', None), **kwargs)
         self.controller = control_interface
         self.nt = 0
-        self.channel_references = {} # reference name -> guild id -> channel name/id
+        self.channel_references = set()
         self.event_listeners = {} # event name -> [listener functions (self, event)]
         # changed to set in favor of event API
         self.event_preemption = {} # event name -> counter for preempting beymax-level events
         self.commands = {} # !cmd -> docstring. Functions take (self, message, content)
-        self.ignored_users = set()
         self.tasks = {} # taskname (auto generated) -> (current exec interval, permanent exec interval)
         self.special = [] # list of (check, handler)
         self.suites = {}
@@ -99,10 +99,10 @@ class Client(discord.Client):
 
     async def async_init(self):
         self.db = DBConnection(
-            await self.controller.config_get('database', 'url')
+            await self.controller.config_get('__global__', 'database', 'url')
         )
     
-    async def get_value(self, key, guild_id=None, default=None):
+    async def get_value(self, key, guild_id=DEFAULT_NULL_GUILD_ID, default=None):
         async with self.db.session() as session:
             result = (await session.execute(
                 alchemy.select(APIEssentials.schema.KeyValue).where(
@@ -114,7 +114,7 @@ class Client(discord.Client):
                 return result.value
             return default
 
-    async def set_value(self, key, value, guild_id=None):
+    async def set_value(self, key, value, guild_id=DEFAULT_NULL_GUILD_ID):
         async with self.db.session() as session:
             result = (await session.execute(
                 alchemy.select(APIEssentials.schema.KeyValue).where(
@@ -158,7 +158,7 @@ class Client(discord.Client):
                 
                 @self.subscribe('ctx:{}'.format(name))
                 async def run_context(bot, event, interaction, *args, **params):
-                    if _command_suite is None or (bot.controller.suite_enabled_in_guild(_command_suite, interaction.guild) and bot.controller.command_enabled_in_guild(name, interaction.guild)):
+                    if _command_suite is None or (bot.controller.suite_enabled_in_guild(_command_suite, interaction.guild) and bot.controller.context_enabled_in_guild(_command_suite, name, interaction.guild)):
                         try:
                             return await func(interaction, *args, **params)
                         except:
@@ -222,7 +222,7 @@ class Client(discord.Client):
 
             @self.subscribe('cmd:{}'.format(cmd.name))
             async def run_cmd(client, event, interaction, **kwargs):
-                if _command_suite is None or (client.controller.suite_enabled_in_guild(_command_suite, interaction.guild) and client.controller.command_enabled_in_guild(command, interaction.guild)):
+                if _command_suite is None or (client.controller.suite_enabled_in_guild(_command_suite, interaction.guild) and client.controller.command_enabled_in_guild(_command_suite, command, interaction.guild)):
                     try:
                         return await cmd._callback(interaction, **kwargs)
                     except:
@@ -357,50 +357,7 @@ class Client(discord.Client):
                 times,
                 '(persistent)' if permanent else '(temporary)'
             )
-        )
-    
-    def add_legacy_task_chain(self, seconds):
-        if seconds < 30:
-            warnings.warn("Legacy task chains cannot guarantee a granularity less than 30 seconds")
-        def wrapper(func):
-            taskname = 'task:'+func.__name__
-            if taskname in self.tasks:
-                raise NameError("This task already exists! Change the name of the task function")
-
-            self.tasks[taskname] = {
-                'seconds': seconds, 
-                'minutes': None,
-                'hours': None,
-                'times': None,
-                'temp_interval': None,
-                'count_remaining': None
-            }
-
-            @self.subscribe('after:schema', once=True)
-            async def verify_task_schedule(bot, *args):
-                async with bot.db.session() as session:
-                    task_schedule = (await session.execute(
-                        alchemy.select(APIEssentials.schema.FutureDispatch).where(
-                            (APIEssentials.schema.FutureDispatch.event == taskname)
-                        )
-                    )).scalars().first()
-                    if task_schedule is None:
-                        logger.info("Scheduling future task chain {}".format(taskname))
-                        await bot.dispatch_future(
-                            seconds,
-                            taskname
-                        )
-            
-            @self.subscribe(taskname) # This is separated into runtask->task only for triggers/hooks on the task event
-            async def task_wrapper(*args):
-                logger.debug("Starting task {}".format(taskname))
-                try:
-                    await func(self)
-                    logger.debug("Task complete {}".format(taskname))
-                finally:
-                    await verify_task_schedule(self)
-        return wrapper   
-            
+        )  
 
     def add_special(self, check, flavor='blind', _command_suite=None): #decorator. Sets the decorated function to run whenever the check is true
         """
@@ -442,7 +399,7 @@ class Client(discord.Client):
             return func
         return wrapper
 
-    async def dispatch_future(self, when, event, *args, **kwargs):
+    async def dispatch_future(self, when, event, guild_context, *args, **kwargs):
         """
         Schedule the given event to be dispatched at a time in the future.
         When can be a (timezone aware) datetime object, timedelta object, or integer (interpreted as seconds in the future).
@@ -467,6 +424,7 @@ class Client(discord.Client):
                     APIEssentials.schema.FutureDispatch(
                         date=when.timestamp(),
                         event=event,
+                        guild_id=guild_context.id if guild_context is not None else DEFAULT_NULL_GUILD_ID,
                         pickled_args=base64.b64encode(pickle.dumps(args)).decode(),
                         pickled_kwargs=base64.b64encode(pickle.dumps(kwargs)).decode(),
                     )
@@ -497,6 +455,13 @@ class Client(discord.Client):
                         )
                     )).scalars().first()
                     if result is None:
+                        try:
+                            session.add_all([
+                                APIEssentials.schema.Migration(key=key, date=discord.utils.utcnow())
+                            ])
+                            await session.commit()
+                        except:
+                            logger.warning("Failed to insert migration key in database (another shard may have already started this migration). Moving on", exc_info=True)
                         logger.info("Running migration {}".format(key))
                         DBView.migration_state.set(True)
                         try:
@@ -508,10 +473,7 @@ class Client(discord.Client):
                             raise
                         finally:
                             DBView.migration_state.set(False)
-                        session.add_all([
-                            APIEssentials.schema.Migration(key=key, date=discord.utils.utcnow())
-                        ])
-                        await session.commit()
+                        
         return wrapper
 
     def subscribe(self, event, *, condition=None, once=False): # decorator. Sets the decorated function to run on events
@@ -583,20 +545,35 @@ class Client(discord.Client):
         """
         # creates a channel reference by that name
         # channel references can be changed in configuration
-        self.channel_references[name] = {}
+        self.channel_references.add(name)
 
-    def fetch_channel(self, name, guild_id):
+    async def fetch_channel_reference(self, name, guild_id):
         """
         Fetch the channel object for a given reference name. If the reference is
         undefined, it returns general
         Arguments:
         name : A string channel reference to lookup
         """
-        channel = None
-        if name in self.channel_references and guild_id in self.channel_references[name]:
-            channel = self.channel_references[name][guild_id]
+        channel_id = await self.controller.config_get(guild_id, 'channels', name, default=None)
+        if channel_id is not None:
+            channel = discord.utils.get(
+                self.get_guild(guild_id).channels,
+                id=int(channel_id)
+            )
+            if channel is not None:
+                return channel
+        # Fallback no1: Try to find general channel
+        channel = discord.utils.get(
+            self.get_guild(guild_id).channels,
+            name='general'
+        )
         if channel is None:
-            return self.channel_references[FALLBACK_GENERAL][guild_id]
+            # Fallback no2: Find any text channel
+            logger.warning("Failed lookup of channel {}. Picking any text channel".format(name))
+            return discord.utils.get(
+                self.get_guild(guild_id).channels,
+                type=discord.ChannelType.text # Pick a random text channel or None. At that point, I'm willing to take the L
+            )
         return channel
 
     def attachSuites(self, *suites):
@@ -711,7 +688,7 @@ class Client(discord.Client):
             )
         return traceID
 
-    async def shutdown(self, signal=None, event=None):
+    async def shutdown(self, signal=None):
         """
         Coroutine. Use this function for a clean shutdown.
         Dispatches the 'cleanup' event, waits for all tasks to complete, then disconnects
@@ -737,9 +714,9 @@ class Client(discord.Client):
         except:
             logger.error("Client shutdown encountered error", exc_info=True)
         finally:
-            if event is not None:
+            if self.controller.shutdown is not None:
                 # Signal the controller that the client has shut down
-                event.set()
+                self.controller.shutdown.set()
 
     async def send_rich_message(self, destination, *, content=None, author=None, author_url=None, author_icon_url=None, title=None, description=None, colour=None, footer=None, image=None, thumbnail=None, __video=None, url=None, reference=None, mention_author=True):
         """
@@ -834,48 +811,7 @@ class Client(discord.Client):
             if result.id != username and '#' not in username:
                 raise NameError("Username '%s' not valid, must containe #discriminator" % username)
             return result.id
-        raise NameError("Unable to locate member '%s'. Must use a user ID, username, or username#discriminator" % username)
-
-    # FIXME: Move to subscription
-    async def on_message(self, message):
-        """
-        Coroutine. Default handler for incomming messages. Do not override.
-        Immediately skips message handling and returns if:
-        * The message was sent by this bot
-        * The message was sent in a DM by a user who does not have any guilds in common with this bot
-        * The message was sent by a user in this bot's ignore list
-
-        Splits the message content by whitespace (or the shlex parser if enabled)
-
-        If the first word starts with the command prefix and is in the list of registered
-        commands, dispatch the command handler, which checks permissions then runs the command
-
-        Otherwise, check if any registered special functions should run on this message
-
-        If you wish to add additional handling for messages, use @bot.subscribe('message').
-        """
-        if message.author == self.user:
-            return
-        if self.get_user(message.author.id) is None:
-            #User is not a member of any known guild
-            #silently ignore
-            return
-
-        if message.author.id in self.ignored_users:
-            logger.debug("Ignoring message from ignored user {}".format(message.author))
-            return
-        # build the user struct and update the users object
-        try:
-            content = message.content.strip().split()
-            content[0] = content[0].lower()
-        except:
-            return
-        # Check if any of the special functions
-        # would like to run on this message
-        for check, event in self.special:
-            if check(self, message):
-                self.dispatch(event, message)
-                break
+        raise NameError("Unable to locate member '%s'. Must use a user ID, username, or username#discriminator" % username)    
 
     def wait_for(self, event, *, check=None, timeout=None):
         """
@@ -915,10 +851,12 @@ APIEssentials = CommandSuite('Beymax Core API Essentials', force_enable=True)
 async def check_future_dispatch(self):
     async with future_dispatch_lock:
         now = discord.utils.utcnow()
+        guild_set = {guild.id for guild in self.guilds} | {DEFAULT_NULL_GUILD_ID, None} # Add None for events still pending from pre-migration
         async with self.db.session() as session:
             results = await session.execute(
                 alchemy.select(APIEssentials.schema.FutureDispatch).where(
-                    APIEssentials.schema.FutureDispatch.date <= now.timestamp()
+                    (APIEssentials.schema.FutureDispatch.date <= now.timestamp())
+                    & (APIEssentials.schema.FutureDispatch.guild_id.in_(guild_set))
                 )
             )
             for event in results.scalars():
@@ -928,20 +866,26 @@ async def check_future_dispatch(self):
                         event.event,
                         overshoot
                     ))
-                await session.delete(event)
-                
+
                 try:
-                    self.dispatch(
-                        event.event,
-                        *pickle.loads(base64.b64decode(as_bytes(event.pickled_args))),
-                        **pickle.loads(base64.b64decode(as_bytes(event.pickled_kwargs)))
-                    )
+
+                    await session.delete(event)
+                
+                    try:
+                        self.dispatch(
+                            event.event,
+                            *pickle.loads(base64.b64decode(as_bytes(event.pickled_args))),
+                            **pickle.loads(base64.b64decode(as_bytes(event.pickled_kwargs)))
+                        )
+                    except:
+                        logger.error("Unable to load event from future dispatch (data type error?). Please check your database", exc_info=True)
                 except:
-                    logger.error("Unable to load event from future dispatch (data type error?). Please check your database", exc_info=True)
+                    logger.info("Unable to delete event {} from database, may have conflicted dispatch. Ignoring event, moving on".format(event.event))
             
             results = (await session.execute(
                 alchemy.select(APIEssentials.schema.FutureDispatch).where(
-                    APIEssentials.schema.FutureDispatch.date > now.timestamp()
+                    (APIEssentials.schema.FutureDispatch.date > now.timestamp())
+                    & (APIEssentials.schema.FutureDispatch.guild_id.in_(guild_set))
                 )
             )).scalars().all()
             await session.commit() # Remove the event even if dispatch fails
@@ -969,24 +913,6 @@ async def first_ready(self, event):
             type=discord.ChannelType.text
         )
         
-        config_channels = await self.controller.config_get('channels', default={}) # Channel name : guild id : channel id
-        self.channel_references[FALLBACK_GENERAL] = {} # Set up fallback general
-        for guild in self.guilds:
-            channels = guild.text_channels
-            general = discord.utils.get(channels, name='general')
-            if general is None:
-                logging.warning("Unable to find general channel in guild {}. Defaulting to first channel present".format(guild.name))
-                general = channels[0]
-            for channel_name, guild_defs in config_channels.items():
-                if channel_name not in self.channel_references:
-                    self.channel_references[channel_name] = {}
-                self.channel_references[channel_name][guild.id] = discord.utils.get(
-                    channels,
-                    id=guild_defs[str(guild.id)]
-                ) if str(guild.id) in guild_defs else None
-                if self.channel_references[channel_name][guild.id] is None:
-                    logger.warning("Channel reference {} not defined or invalid ID for guild {}".format(channel_name, guild.name))
-        
         await (self.dispatch('schema', self.db)) # Returns a task
 
         # Add legacy lock
@@ -1000,6 +926,8 @@ async def first_ready(self, event):
             logger.error("Migrations timed out after 3 minutes")
         except:
             logger.error("Errors encountered during migrations", exc_info=True)
+        
+        logger.info("Migrations complete")
 
         for task in self.tasks.values():
             if 'task' in task: # Legacy task chains do not set a task attribute
@@ -1038,46 +966,6 @@ async def sync_commands(self, evt):
         await self.tree.sync(guild=guild) # Clear out guild commands. We're global-only, baby
     await self.change_presence()
 
-@APIEssentials.subscribe('guild_role_create')
-@APIEssentials.subscribe('guild_role_delete')
-@APIEssentials.subscribe('before:ready', once=True)
-async def update_ignore_roles(self, event, *args):
-    """
-    Updates Beymax's list of roles to ignore
-    """
-    references = await self.controller.config_get('ignore_roles', default=[])
-    if not isinstance(references, list):
-        references = [references]
-    self.ignore_roles = {
-        role
-        for guild in self.guilds
-        for role in guild.roles
-        if role.name in references or role.id in references
-    }
-    logger.debug("Populated {} roles to ignore".format(len(self.ignore_roles)))
-    if len(references) > 0 and len(self.ignore_roles) == 0:
-        warnings.warn(
-            "No ignore roles matched. There is currently no way to blacklist users"
-        )
-        logger.warning("No ignore roles matched. There is currently no way to blacklist users")
-    self.ignored_users = set()
-    for member in self.get_all_members():
-        if len(set(member.roles) & self.ignore_roles):
-            logger.debug("Ignoring user {}".format(member))
-            self.ignored_users.add(member.id)
-
-@APIEssentials.subscribe('after:member_update') # Event not delivered 
-async def update_ignore_users(self, event, before, after):
-    """
-    Updates Beymax's list of users to ignore
-    """
-    if len(set(after.roles) & self.ignore_roles):
-        logger.debug("Ignoring user {}".format(after))
-        self.ignored_users.add(after.id)
-    elif after.id in self.ignored_users:
-        logger.debug("Unignoring user {}".format(after))
-        self.ignored_users.remove(after.id)
-
 @APIEssentials.add_command('parse', description="Parse this string and run special message handlers")
 async def cmd_parse(interaction, message: Arg(description="Message to pass to the bot", type=str)):
     msg = lambda : None # Dummy namespace object
@@ -1094,9 +982,10 @@ class Migration(object):
     key = alchemy.Column(alchemy.String, primary_key=True)
     date = alchemy.Column(alchemy.DateTime(timezone=True))
 
-@APIEssentials.global_table
+@APIEssentials.table
 class FutureDispatch(object):
     __tablename__ = 'core_future_dispatch'
+    guild_id = alchemy.Column(alchemy.BigInteger)
     date = alchemy.Column(alchemy.Integer, primary_key=True)
     event = alchemy.Column(alchemy.String, primary_key=True)
     pickled_args = alchemy.Column(alchemy.TEXT)
@@ -1106,7 +995,7 @@ class FutureDispatch(object):
 class KeyValue(object):
     __tablename__ = 'core_key_value'
     key = alchemy.Column(alchemy.String, primary_key=True)
-    guild_id = alchemy.Column(alchemy.BigInteger)
+    guild_id = alchemy.Column(alchemy.BigInteger, primary_key=True)
     value = alchemy.Column(alchemy.String)
 
 @APIEssentials.global_table
@@ -1140,6 +1029,7 @@ async def migrate_db_backend(bot):
                 APIEssentials.schema.FutureDispatch(
                     date=strptime(event['date']).timestamp(),
                     event=event['event'],
+                    guild_id=DEFAULT_NULL_GUILD_ID,
                     pickled_args=base64.b64encode(pickle.dumps(event['args'])).decode(),
                     pickled_kwargs=base64.b64encode(pickle.dumps(event['kwargs'])).decode()
                 )
@@ -1169,6 +1059,67 @@ async def add_guild_id(client):
         except:
             logger.warning("Unable to complete add-guild-id migration -- assuming correct schema")
 
+@APIEssentials.migration('future-dispatch-guild-id')
+async def add_guild_id(client):
+    async with client.db.session() as session:
+        try:
+            await session.execute(
+                "ALTER TABLE core_future_dispatch ADD COLUMN guild_id {};".format(
+                    alchemy.BigInteger().compile(client.db.engine.dialect)
+                )
+            )
+            await session.commit()
+        except:
+            logger.warning("Unable to complete add-guild-id migration -- assuming correct schema")
+
+@APIEssentials.migration('make-guild-id-pkey')
+async def add_guild_id(client):
+    async with client.db.session() as session:
+        try:
+            await session.execute(
+                "UPDATE core_key_value SET guild_id = {} WHERE guild_id IS null;".format(
+                    DEFAULT_NULL_GUILD_ID
+                )
+            )
+            await session.execute(
+                "ALTER TABLE core_key_value DROP CONSTRAINT core_key_value_pkey CASCADE, ADD PRIMARY KEY (key, guild_id);"
+            )
+            await session.commit()
+        except:
+            logger.warning("Unable to complete add-guild-id migration -- assuming correct schema")
+
 @APIEssentials.subscribe('guild_join')
 async def on_guild_join(client, event, guild):
     await client.controller.on_guild_join(guild)
+
+@APIEssentials.subscribe('message')
+async def special_message_handler(client, event, message):
+        """
+        Coroutine. Default handler for incomming messages. Do not override.
+        Immediately skips message handling and returns if:
+        * The message was sent by this bot
+        * The message was sent in a DM by a user who does not have any guilds in common with this bot
+        * The message was sent by a user in this bot's ignore list
+
+        Splits the message content by whitespace (or the shlex parser if enabled)
+
+        If the first word starts with the command prefix and is in the list of registered
+        commands, dispatch the command handler, which checks permissions then runs the command
+
+        Otherwise, check if any registered special functions should run on this message
+
+        If you wish to add additional handling for messages, use @bot.subscribe('message').
+        """
+        if message.author == client.user:
+            return
+        
+        
+        if client.user.id in {mention.id for mention in message.mentions}:
+            client.dispatch('mention', message)
+
+        # Check if any of the special functions
+        # would like to run on this message
+        for check, special_event in client.special:
+            if check(client, message):
+                client.dispatch(special_event, message)
+                break
