@@ -23,6 +23,7 @@ class CommandSuite(object):
         self.context = []
         self.schema = Schema()
         self.dependencies = []
+        self.dependents = []
         self.tables = []
         self.command_group = None
         self.forced = force_enable
@@ -43,6 +44,7 @@ class CommandSuite(object):
     
     def add_dependency(self, suite):
         self.dependencies.append(suite)
+        suite.dependents.append(self)
 
     def table(self, cls):
         """
@@ -56,6 +58,7 @@ class CommandSuite(object):
         @self.subscribe('schema', once=True)
         async def add_table(bot, evt, db):
             class Table(db.base, cls):
+                _table_original_name = cls.__name__
                 _table_guild_scope = True
                 # if hasattr(cls, '_relationships_'):
                 #     for attr, val in cls._relationships_.items():
@@ -78,6 +81,7 @@ class CommandSuite(object):
         @self.subscribe('schema', once=True)
         async def add_table(bot, evt, db):
             class Table(db.base, cls):
+                _table_original_name = cls.__name__
                 _table_guild_scope = False
                 # if hasattr(cls, '_relationships_'):
                 #     for attr, val in cls._relationships_.items():
@@ -136,22 +140,17 @@ class CommandSuite(object):
             bot.add_special(special['checker'], flavor=special['flavor'])(special['function'])
 
         for task in self.tasks:
-            if task['legacy']:
-                bot.add_legacy_task_chain(
-                    task['seconds']
-                )(task['function'])
-            else:
-                bot.add_task(
-                    task['seconds'], minutes=task['minutes'], 
-                    hours=task['hours'], times=task['times'], 
-                    count=task['count'], reconnect=task['reconnect']
-                )(task['function'])
+            bot.add_task(
+                task['seconds'], minutes=task['minutes'], 
+                hours=task['hours'], times=task['times'], 
+                count=task['count'], reconnect=task['reconnect']
+            )(task['function'])
 
         for command in self.commands:
-            bot.add_command(command['command'], *command['args'], command_suite_group=self.command_group, **command['kwargs'])(command['function'])
+            bot.add_command(command['command'], *command['args'], command_suite_group=self.command_group, _command_suite=self, **command['kwargs'])(command['function'])
         
         for ctx in self.context:
-            bot.add_context(ctx['name'], *ctx['args'], **ctx['kwargs'])(ctx['function'])
+            bot.add_context(ctx['name'], *ctx['args'], _command_suite=self, **ctx['kwargs'])(ctx['function'])
         
         if self.command_group is not None:
             @bot.subscribe('before:command-init', once=True)
@@ -185,9 +184,7 @@ class CommandSuite(object):
             return func
         return wrapper
 
-    def add_task(self, seconds=None, *, minutes=None, hours=None, times=None, count=None, reconnect=None, legacy=False):
-        if legacy and not (minutes is None and hours is None and times is None and count is None and reconnect is None):
-            raise ValueError("Cannot create legacy task chain with complex scheduling argumentsß")
+    def add_task(self, seconds=None, *, minutes=None, hours=None, times=None, count=None, reconnect=None):
         def wrapper(func):
             self.tasks.append(
                 {
@@ -198,7 +195,6 @@ class CommandSuite(object):
                     'count': count,
                     'reconnect': reconnect,
                     'function': func,
-                    'legacy': legacy
                 }
             )
             return func
